/*
** Author: Dimitris Diochnos
** email : diochnos AT gmail DOT com
**
** File Name: mighty_classes.cpp
**
** Date Started : 20 April 2006.
** Date Finished: 28 April 2006.
**
** internals_mc (version 1.0)
** Date Started :  6 July 2007.
** Date Finished: 11 July 2007.
**
** internals_mc (version 1.1)
** Date Started : 12 July 2007.
** Date Finished: 20 July 2007.
**
**
** Tab space: 3
**
**
** The "main" file.
**
*/

// So that you avoid warning for strcpy () in Visual Studio:
#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE 1
#endif

#include <fstream>
#include <iostream>
#include <iomanip>
#include <cstring>
#include <ctime>
#include <cctype>
#include "mighty_classes.h"
#include "sec_skills_mighty.h"
//#include "math_toolbox.h"

#define CURRENT_VERSION_NUM 1.1
#define CURRENT_VERSION_STR "1.1"

/*
** External global variables.
*/
extern char * Alchemists_Names [];
extern char * Barbarians_Names [];
extern char * Beastmasters_Names [];
extern char * Death_Knights_Names [];
extern char * Demoniacs_Names [];
extern char * Knights_Names [];
extern char * Overlords_Names [];
extern char * Planeswalkers_Names [];
extern char * Rangers_Names [];

/*****************************************************************************/

unsigned char * Read_Selections (char *);
bool 	are_valid (HERO_PTR, unsigned char *, char *);
void default_version_message ();
void default_about_message ();
void default_error_message ();
void default_help_message ();
void default_log_message ();
void PrintSelections (unsigned char *);

/*****************************************************************************/

int 	main (int argc, char * argv [])
{
	clock_t 			start, finish;
	unsigned char 	pointer;
	long unsigned int episodes;
	int 					algorithm;
	unsigned char * selections;
	DecisionMaker * dMaker;

	if (argc == 2) {
		char * request = new char [strlen(argv[1]) + 1];

		strcpy(request, argv[1]);
		for (pointer = 0; pointer < strlen (argv [1]); pointer++)
			request [pointer] = toupper ( request [pointer] );

		if (strcmp(request, "-VERSION") == 0)
			default_version_message ();
		else if (strcmp(request, "-HELP") == 0)
			default_help_message ();
		else if (strcmp(request, "-ABOUT") == 0)
			default_about_message ();
		else if (strcmp(request, "-LOG") == 0)
			default_log_message ();
		else {
			default_error_message ();

			exit(1);
		}

		delete [] request;

		exit(0);
	}
	if ((argc != 5) && (argc != 6)) {
		default_error_message ();

		exit (1);
	}


	/*
	** Program has 4 or 5 arguments. Prepare for execution.
	*/


	char * 	given_class = new char [strlen (argv [1]) + 1];
	char * 	given_name = new char [strlen (argv [2]) + 1];

	// Convert to UPPERCASE.
	strcpy (given_class, argv [1]);
	for (pointer = 0; pointer < strlen (argv [1]); pointer++)
		given_class [pointer] = toupper ( given_class [pointer] );

	// Convert to UPPERCASE.
	strcpy (given_name, argv [2]);
	for (pointer = 0; pointer < strlen (argv [2]); pointer++)
		given_name [pointer] = toupper ( given_name [pointer] );

	/*
	** Validate user input.
	*/
	unsigned char 	classification;
	unsigned char 	name;

	if (strcmp (given_class, "ALCHEMIST") == 0 )
		classification = ALCHEMIST;
	else if (strcmp (given_class, "BARBARIAN") == 0 )
		classification = BARBARIAN;
	else if (strcmp (given_class, "BEASTMASTER") == 0 )
		classification = BEASTMASTER;
	else if (strcmp (given_class, "DEATH_KNIGHT") == 0 )
		classification = DEATH_KNIGHT;
	else if (strcmp (given_class, "DEMONIAC") == 0 )
		classification = DEMONIAC;
	else if (strcmp (given_class, "KNIGHT") == 0 )
		classification = KNIGHT;
	else if (strcmp (given_class, "OVERLORD") == 0 )
		classification = OVERLORD;
	else if (strcmp (given_class, "PLANESWALKER") == 0 )
		classification = PLANESWALKER;
	else if (strcmp (given_class, "RANGER") == 0 )
		classification = RANGER;
	else {
		// Free-up memory for first two arguments.
		delete 	[] given_class;
		delete 	[] given_name;

		std::cerr << "Invalid Hero class given." << std::endl;
		std::cerr << "Please execute program with -help parameter to view help." << std::endl;

		exit (2);
	}

	switch(classification) {
		case ALCHEMIST:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Alchemists_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case BARBARIAN:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Barbarians_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case BEASTMASTER:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Beastmasters_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case DEATH_KNIGHT:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Death_Knights_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case DEMONIAC:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Demoniacs_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case KNIGHT:
			for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) {
				if (strcmp (given_name, Knights_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case OVERLORD:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Overlords_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case PLANESWALKER:
			for (pointer = 0; pointer < NUM_HEROES; pointer++) {
				if (strcmp (given_name, Planeswalkers_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		case RANGER:
			for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) {
				if (strcmp (given_name, Rangers_Names [pointer]) == 0) {
					name = pointer;
					goto 	normal_execution;
				}
			}
			goto 	name_error;
			break;

		default:
name_error:
			std::cerr << "Invalid Hero name given." << std::endl;
			std::cerr << "Please execute program with -help parameter to view help." << std::endl;

			// Free-up memory for first two arguments.
			delete 	[] given_class;
			delete 	[] given_name;

			exit (4);

			break;
	}


	/*
	** Procceed to execution of Always Left policy.
	*/
normal_execution:

	// Free-up memory for first two arguments.
	delete 	[] given_class;
	delete 	[] given_name;

	char * 	given_algorithm = new char [strlen (argv [3]) + 1];

	strcpy (given_algorithm, argv [3]);
	for (pointer = 0; pointer < strlen (argv [3]); pointer++)
		given_algorithm [pointer] = toupper ( given_algorithm [pointer] );

	if (strcmp (given_algorithm, "AR") == 0 )
		algorithm = AR_ALGO;
	else if (strcmp (given_algorithm, "ANSA") == 0 )
		algorithm = ANSA_ALGO;
	else if (strcmp (given_algorithm, "AL") == 0 )
		algorithm = AL_ALGO;
	else if (strcmp (given_algorithm, "ALTP") == 0 )
		algorithm = ALTP_ALGO;
	else if (strcmp (given_algorithm, "SPOU") == 0 )
		algorithm = SPOU_ALGO;
	//else if (strcmp (given_algorithm, "FMTP") == 0 )
	//	algorithm = FMTP_ALGO;
	else {
		delete 	[] given_algorithm;

		std::cerr << "Invalid algorithm given." << std::endl;
		std::cerr << "Please execute program with -help parameter to view help." << std::endl;

		exit (5);
	}

	// Free-up memory.
	delete 	[] given_algorithm;



	// Create an instance of the selected hero.
	HERO_PTR	yet_another_hero = new HERO (classification, name);




	// Initialize DecisionMaker based on algorithm selection.
	switch(algorithm) {
		//case AR_ALGO:
		case ANSA_ALGO:
			dMaker = new HeroAlwaysRight (yet_another_hero);
			break;

		case AL_ALGO:
			dMaker = new HeroAlwaysLeft (yet_another_hero);
			break;

		case ALTP_ALGO:
			selections = Read_Selections ( DEFAULT_ALTP_PREFS_FILE );
			if (are_valid (yet_another_hero, selections, DEFAULT_ALTP_PREFS_FILE))
				dMaker = new HeroAlwaysLeftThenPref (yet_another_hero, selections);
			else {
				delete 	[] selections;
				exit(6);
			}
			delete 	[] selections;
			break;

		case SPOU_ALGO:
			selections = Read_Selections ( DEFAULT_SPOU_PREFS_FILE );
			//std::cout << "about to print selections" << std::endl;
			//PrintSelections (selections);
			//std::cout << "selections printed" << std::endl;
			if (are_valid (yet_another_hero, selections, DEFAULT_SPOU_PREFS_FILE)) {
				dMaker = new HeroSeekPrefOtherwiseUpgrade (yet_another_hero, selections);
				//dynamic_cast<HeroSeekPrefOtherwiseUpgrade *>(dMaker)->printPreferences ();
			}
			else {
				delete 	[] selections;
				exit(6);
			}
			delete 	[] selections;
			break;

//		case FMTP_ALGO:
//			std::cerr << "The selected algorithm is not supported yet!" << std::endl;
//			exit(7);
			//dMaker = new HeroAlwaysLeft (yet_another_hero);
//			break;

		default:
			std::cerr << "You should not enter here while selecting policy!" << std::endl;
			exit(7);
			break;
	}



	/*
	** Fetch number of episodes.
	*/
	episodes = (long unsigned int) atoi( argv[4] );




	/*
	** Fetch printing procedure.
	** By default use FANCY output.
	*/
	unsigned char	output_algo = FANCY;
	if (argc == 6) {
		char * requested_output = new char [strlen(argv[5]) + 1];
		strcpy (requested_output, argv[5]);
		for (pointer = 0; pointer < strlen(argv[5]) + 1; pointer++)
			requested_output[pointer] = toupper( requested_output[pointer] );		// Convert to uppercase.

		if (strcmp(requested_output, "DEPRECATED") == 0)
			output_algo = DEPRECATED;
		else if (strcmp(requested_output, "NEAT") == 0)
			output_algo = NEAT;
		else if (strcmp(requested_output, "FANCY") != 0)
			std::cerr << "Warning: Invalid output requested. FANCY used instead." << std::endl;

		delete	[] requested_output;
	}


	// Make it "truly random".
	unsigned int seed;
	seed = (unsigned int) time ( NULL );
	srand( seed );
	//std::cout << "Seed        : " << seed << std::endl;


	// Store user preferences for number of episodes and printing method.
	// Also store seed just in case it is needed for debugging within the
	// Math Toolbox.
	MT_PTR myMT = new MT;

	myMT->setNumberOfEpisodes ( episodes );
	if (output_algo != FANCY)
		myMT->setPrintingMethod ( output_algo );
	myMT->setSeed ( seed );


	/*
	** Apply MonteCarlo with selected options.
	*/
	MonteCarlo * MC = new MonteCarlo ( yet_another_hero, dMaker );

	// Count!
	start = clock ();
	//std::cout << "About to run MonteCarlo." << std::endl;
	MC->run ( myMT );
	//std::cout << "MonteCarlo finished." << std::endl;
	finish = clock ();



	/*
	** Print Results for specific hero.
	*/
	if ((algorithm == ANSA_ALGO) || (algorithm == AL_ALGO)) {
		if (episodes >= 400000000)
			std::cerr << "\aPossible overflow error while smoothing results." << std::endl;
		myMT->smoothResults (yet_another_hero);	// Smooth Results if necessary.
	}
	myMT->printStatistics (yet_another_hero, start, finish); 


	// Free-up memory reserved.
	delete	yet_another_hero;
	delete	dMaker;
	delete	MC;
	delete	myMT;

	//char tempoulas[32];
	//std::cin.getline (tempoulas, 31);


	return 	0;
}

/*****************************************************************************/

unsigned char * Read_Selections (char * name)
{
	int i, j;
	char current_line [14];
	bool already_found[DIFF_SEC_SKILLS];
	unsigned char * selections = (unsigned char *) NULL;

	/* Check if file 'selections.txt' exists. */
	std::ifstream selFD ( name );

	if ( !selFD ) {
		std::cerr << "Usage: Program expects file " << name << " in same directory.\n";

		exit (6);
	}

	// Initialize "double_appearence" array.
	for (i = 0; i < DIFF_SEC_SKILLS; i++)
		already_found[i] = false;	// No skill is already read.

	/*
	** Store user's selections in memory.
	*/
	selections = new unsigned char [DIFF_SEC_SKILLS];
	
	for (i = 0; i < (DIFF_SEC_SKILLS - 1); i++) {
		selFD.getline (current_line, 13);
		for (j = 0; j < (int) strlen (current_line ); j++) //convert selection to uppercase
				current_line[j] = (char) toupper( (int) current_line[j] );
		for (j = 1; j < DIFF_SEC_SKILLS; j++) {
			if ( strcmp( current_line, get_skill_name_consecutive (j) ) == 0 ) {
				if ( !already_found[j] )
					already_found [j] = true;
				else {
					// We have a problem because same skill was read earlier in file.
					// Indicate error to user.
					std::cerr << name << ": Duplicate skill (" << get_skill_name_consecutive (j);
					std::cerr << ") on input line " << (int) (i + 1) << "." << std::endl;

					exit(6);
				}
				selections[i] = j; 	// Store preference.
				break;
			}
		}
		if ( j == DIFF_SEC_SKILLS ) {
			// Parser was unable to locate skill given by user.
			std::cerr << name << ": Invalid skill given on input line " << (int) (i + 1) << "." << std::endl;

			exit (6);
		}
		//std::cout << get_skill_name_consecutive(k) << "." << std::endl;
	}
	selections[DIFF_SEC_SKILLS - 1] = NULL_SKILL; 	// Null-skill no-one wants it.

	selFD.close ();

	return selections;
}

/*****************************************************************************/

bool 	are_valid (HERO_PTR yet_another_hero, unsigned char * selections, char * filename)
{
	int i;
	bool found;
	//char temp[32];

	found = false;
	//std::cout << "selections[0] : " << get_skill_name (selections[0]) << std::endl;
	//std::cout << "slot      [0] : " << get_skill_name (yet_another_hero->get_skill_from_slot (0)) << std::endl;
	for (i = 0; i < SKILL_SLOTS; i++) {
		if (selections[i] == yet_another_hero->get_skill_from_slot (0)) {
			found = true;
			break;
		}
	}
	if (!found) {
		std::cerr << "Hero starts with " << get_skill_name( yet_another_hero->get_skill_from_slot (0) );
		std::cerr << " and you did not provide it in the first 8 skills." << std::endl;
		std::cerr << "Please edit your selections in " << filename << " file." << std::endl;

		//std::cin.getline ( temp, 31 );

		return false;;
	}

	if (yet_another_hero->get_skill_from_slot(1) != NULL_SKILL) {
		found = false;
		for (i = 0; i < SKILL_SLOTS; i++) {
			if (selections[i] == yet_another_hero->get_skill_from_slot (1)) {
				found = true;
				break;
			}
		}

		if (!found)	{
			std::cerr << "Hero starts with " << get_skill_name( yet_another_hero->get_skill_from_slot (1) );
			std::cerr << " and you did not provide it in the first 8 skills." << std::endl;
			std::cerr << "Please edit your selections in " << filename << " file." << std::endl;

			//std::cin.getline ( temp, 31 );

			return false;
		}
	}

	return true;
}

/*****************************************************************************/

void default_version_message ()
{
	std::cout << CURRENT_VERSION_STR << std::endl;
}

void default_about_message ()
{
	std::cout << "internals_mc: about" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~~~" << std::endl;
	std::cout << "Version: " << CURRENT_VERSION_STR << std::endl;
	std::cout << "Release: July 20, 2007" << std::endl;
	std::cout << "* For a complete documentation and comments please visit homepage:" << std::endl;
	std::cout << "  http://heroescommunity.com/viewthread.php3?TID=17812&pagenumber=8" << std::endl;
	std::cout << "* Bugs and bugfixes can be reported in the forum above (recommended)" << std::endl;
	std::cout << "  or with a direct mail at: diochnos AT gmail DOT com." << std::endl;
}

void default_error_message ()
{
	std::cout << "internals_mc" << std::endl;
	std::cout << "~~~~~~~~~~~~" << std::endl;
	std::cout << "Incorrect parameters. Please execute:" << std::endl;
	std::cout << "internals_mc -help" << std::endl;
	std::cout << "for a complete list of available options or visit homepage at:" << std::endl;
	std::cout << "http://heroescommunity.com/viewthread.php3?TID=17812&pagenumber=8" << std::endl;
}

void default_help_message ()
{
	int pointer;

	std::cout << "internals_mc: help" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~~" << std::endl;
	std::cout << "Usage: Program expects 4 or 5 arguments." << std::endl;
	std::cout << "1st  : Hero Class." << std::endl;
	std::cout << "2nd  : Hero Name." << std::endl;
	std::cout << "3rd  : Policy (Skill Selection Algorithm)." << std::endl;
	std::cout << "4th  : Number of episodes." << std::endl;
	std::cout << "5th  : Printing method (optional argument)." << std::endl << std::endl;
	std::cout << "Valid names for Hero Classes:" << std::endl;
	std::cout << "Alchemist, Barbarian, Beastmaster, Death_Knight, " << std::endl;
	std::cout << "Demoniac, Knight, Overlord, Planeswalker, Ranger." << std::endl << std::endl;
	std::cout << "Valid names for Heroes:" << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Alchemists_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Barbarians_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Beastmasters_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Death_Knights_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Demoniacs_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) { 	// There is also Sir Mullich.
		std::cout << Knights_Names [pointer] << " ";
		if (pointer == 4)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Overlords_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES; pointer++) {
		std::cout << Planeswalkers_Names [pointer] << " ";
		if (pointer == 3)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl;
	for (pointer = 0; pointer < NUM_HEROES + 1; pointer++) { 	// There is also Gelu.
		std::cout << Rangers_Names [pointer] << " ";
		if (pointer == 4)
			std::cout << std::endl;
	}
	std::cout << std::endl << std::endl << std::endl;
	std::cout << "Valid names for algorithms (policies):" << std::endl;
	std::cout << "AR  : Always Right." << std::endl;
	std::cout << "ANSA: Always Right (same as above)." << std::endl;
	std::cout << "AL  : Always Left." << std::endl;
	std::cout << "ALTP: Always Left-Then Preference." << std::endl;
	std::cout << "SPOU: Seek Preference-Otherwise Upgrade (according to Preference)." << std::endl;
	//std::cout << "FMTP: First Magic-Then Preference" << std::endl;
	std::cout << std::endl << std::endl;

	std::cout << "Valid names for Printing Methods:" << std::endl;
	std::cout << "DEPRECATED: The initial (raw) output." << std::endl;
	std::cout << "            Might be useful in conjuction with 'ansa2excel' program." << std::endl;
	std::cout << "NEAT      : A neat version for the deprecated output." << std::endl;
	std::cout << "FANCY     : Generates FANCY output." << std::endl;
	std::cout << "            This is used by default." << std::endl;

	std::cout << std::endl;
}

void	default_log_message ()
{
	std::cout << "internals_mc: log" << std::endl;
	std::cout << "~~~~~~~~~~~~~~~~~" << std::endl << std::endl;
	std::cout << "Version 1.0:" << std::endl;
	std::cout << " * The initial version of the program." << std::endl;
	std::cout << " - Developed: July 6 - 11, 2007." << std::endl;
	std::cout << " - Released : July 11, 2007." << std::endl;
	std::cout << std::endl;
	std::cout << "Version 1.1:" << std::endl;
	std::cout << " * Fixed bugs when reading preference files for ALTP and SPOU policies." << std::endl;
	std::cout << " * Optimization enhancements." << std::endl;
	std::cout << " * Generation of a Mathematical toolbox that allows:" << std::endl;
	std::cout << "   > Smoothed results for AR and AL policies." << std::endl;
	std::cout << "   > Different printing methods of the results." << std::endl;
	std::cout << " * Fixed bug in SPOU policy behaviour. More info on thread." << std::endl;
	std::cout << " * Allowed the use of this log." << std::endl;
	std::cout << " - Developed: July 12 - 20, 2007." << std::endl;
	std::cout << " - Released : July 20, 2007." << std::endl;
}

void 	PrintSelections(unsigned char * selections)
{
	int i;

	std::cout << "SELECTIONS" << std::endl;
	for (i = 0; i < DIFF_SEC_SKILLS; i++) {
		std::cout << get_skill_name_12 ( i ) << ": " << (int) selections[i] << std::endl;
	}
	std::cout << "----------" << std::endl;
}
