/*
** Author: Dimitris Diochnos
** email : diochnos AT gmail DOT com
**
** File Name: mighty_classes.cpp
**
** Date Started : 20 April 2006.
** Date Finished: 28 April 2006.
**
** internals_mc (version 1.0)
** Date Started :  6 July 2007.
** Date Finished: 11 July 2007.
**
** internals_mc (version 1.1)
** Date Started : 12 July 2007.
** Date Finished: 20 July 2007.
**
**
** Tab space: 3
**
**
** The source file for mathematical toolbox.
**
*/

#include <iostream>
#include <iomanip>
#include "mighty_classes.h"

/*****************************************************************************/

MathToolbox::MathToolbox ()
{
	printing_method = (unsigned char) FANCY;	// By default, FANCY is used.

	episodes = (long unsigned int) 0;

	initializeOccurrences ();
}

MathToolbox::~MathToolbox () {}

/*****************************************************************************/

/*****************************************\
* SETTING AND GETTING THE PRINTING METHOD *
\*****************************************/

unsigned char			MathToolbox::getPrintingMethod () const {	return	printing_method; }

// Set a different PrintingMethod.
bool	MathToolbox::setPrintingMethod (const unsigned char method)
{
	if (method <= (unsigned char) MAX_PRINTING_METHOD) {
		printing_method = method;
		return true;
	}
	else {
		// Invalid method given.
		// No changes made. Returning false.
		return false;
	}
}

/*****************************************************************************/

// Manipulating 'occurrences' array.
bool 						MathToolbox::incrementOccurrences (const unsigned char skill)
{
	if (skill >= DIFF_SEC_SKILLS)
		return false;	// Indicate error.

	occurrences[skill]++;

	return 	true; 	// Indicate correct behaviour.
}


long unsigned int 	MathToolbox::getOccurrences (const unsigned char skill) const
{
	if (skill >= DIFF_SEC_SKILLS)
		return 0;

	return 	occurrences[skill];
}

/*****************************************************************************/

void 						MathToolbox::setNumberOfEpisodes (const long unsigned int number)
{
	episodes = number;
}

long unsigned int 	MathToolbox::getNumberOfEpisodes () const
{
	return 	episodes;
}

/*****************************************************************************/

void 						MathToolbox::setSeed (const unsigned int s)
{
	seed = s;
}


unsigned int 			MathToolbox::getSeed () const
{
	return seed;
}

/*****************************************************************************/

void 						MathToolbox::initializeOccurrences ()
{
	for (int i = 0; i < DIFF_SEC_SKILLS; i++)
		occurrences[i] = 0;
}

/*****************************************************************************/

void 						MathToolbox::setProbabilities ()
{
	for (int i = 0; i < DIFF_SEC_SKILLS; i++)
		probabilities [i] = (double) ((double) occurrences [i]) / ((double) episodes);
}

/*****************************************************************************/

void						MathToolbox::smoothResults (HERO_PTR yah)
{
	int i, weight, size, remainder;
	//bool	belongs_in_group[DIFF_SEC_SKILLS];
	unsigned char group_members [DIFF_SEC_SKILLS];
	long unsigned int sum;
	unsigned char * chances = yah->get_skill_chances ();

	// Smooth everything apart from WISDOM and MAGIC Schools.
	weight = 1;
	while (weight <= 10) {
		size = 0;
		for (i = 0; i < DIFF_SEC_SKILLS; i++) {
			if ((!yah->is_magic_school (i)) && (i != WISDOM) && (chances[i] == weight) && (yah->get_skill_slot (i) >= yah->get_initial_skills ())) {
				// It is neither magic school nor wisdom.
				//belongs_in_group [i] = true;
				group_members [size++] = i;
			}
		}

		if (size == 0) {
			weight++;
			continue;
		}

		// Logically, there will be no more than 10 members in every group.
		// Hence, overflows for episodes >= (2^32 - 1) / 10;
		sum = 0;
		for (i = 0; i < size; i++)
			sum += occurrences [ group_members[i] ];
		remainder = sum % size;
//		std::cout << "Remainder: " << remainder << std::endl;
		for (i = 0; i < size; i++) {
//			std::cout << "Setting " << get_skill_name ( group_members [i] ) << " to: ";
//			std::cout << (long unsigned int) sum / size << "." << std::endl;
			occurrences [ group_members [i] ] = (long unsigned int) sum / size;
		}
		// Add remainder from top to bottom.
		for (i = 0; i < remainder; i++)
			occurrences [ group_members [i] ]++;

		weight++; // Procceed to next group.
	}

	// Now smooth MAGIC Schools.
	weight = 1;
	while (weight <= 10) {
		size = 0;
		for (i = 0; i < DIFF_SEC_SKILLS; i++) {
			if ((yah->is_magic_school (i)) && (chances[i] == weight) && (yah->get_skill_slot (i) >= yah->get_initial_skills ())) {
				// It is a magic school.
				//belongs_in_group [i] = true;
				group_members [size++] = i;
			}
		}

		if (size == 0) {
			weight++;
			continue;
		}

		// Logically, there will be no more than 10 members in every group.
		// Hence, overflows for episodes >= (2^32 - 1) / 10;
		sum = 0;
		for (i = 0; i < size; i++)
			sum += occurrences [ group_members[i] ];
		remainder = sum % size;
//		std::cout << "Remainder: " << remainder << std::endl;
		for (i = 0; i < size; i++) {
//			std::cout << "Setting " << get_skill_name ( group_members [i] ) << " to: ";
//			std::cout << (long unsigned int) sum / size << "." << std::endl;
			occurrences [ group_members [i] ] = (long unsigned int) sum / size;
		}
		// Add remainder from top to bottom.
		for (i = 0; i < remainder; i++)
			occurrences [ group_members [i] ]++;

		weight++; // Procceed to next group.
	}

	delete	[] chances;
}

/*****************************************************************************/

/*********************\
* PRINTING STATISTICS *
\*********************/

void 	MathToolbox::printStatistics (HERO_PTR yah, const clock_t start, const clock_t finish)
{
	setProbabilities ();
	
	if (printing_method == DEPRECATED)
		printStatisticsDeprecated (yah, start, finish);
	else if (printing_method == NEAT)
		printStatisticsNeat (yah, start, finish);
	else
		printStatisticsFancy (yah, start, finish);
}

/*****************************************************************************/

// Private functions called by PrintStatistics.
void MathToolbox::printStatisticsDeprecated (HERO_PTR yah, const clock_t start, const clock_t finish)
{
	unsigned char pointer;

	yah->print_Name ();
	for (pointer = 1; pointer < (unsigned char) DIFF_SEC_SKILLS; pointer++)
		std::cout << (char *) get_skill_name_12 (pointer) << ": " << std::setprecision (4) << (double) (100.0 * probabilities [pointer]) << " %." << std::endl;
	std::cout << std::endl;

	std::cout << "Duration    : " << std::setprecision (2) <<  std::fixed << (long double) (finish - start) / CLOCKS_PER_SEC << " seconds." << std::endl;
	std::cout << "Seed        : " << seed << std::endl;
}

void MathToolbox::printStatisticsNeat (HERO_PTR yah, const clock_t start, const clock_t finish)
{
	unsigned char pointer;

	yah->print_Name ();
	for (pointer = 1; pointer < (unsigned char) DIFF_SEC_SKILLS; pointer++)
		std::cout << (char *) get_skill_name_12 (pointer) << ": " << std::setw (8) << std::setprecision (4) << std::fixed << (double) (100.0 * probabilities [pointer]) << " %." << std::endl;
	std::cout << std::endl;

	std::cout << "Duration    : " << std::setprecision (2) <<  std::fixed << (long double) (finish - start) / CLOCKS_PER_SEC << " seconds." << std::endl;
	std::cout << "Seed        : " << seed << std::endl;
}

void MathToolbox::printStatisticsFancy (HERO_PTR yah, const clock_t start, const clock_t finish)
{
	unsigned char	pointer, j, signs;

	yah->print_Name ();
	for (pointer = 1; pointer < (unsigned char) DIFF_SEC_SKILLS; pointer++) {
		std::cout << (char *) get_skill_name_12 (pointer) << " |";
		if ((probabilities[pointer] > 0.0) && (probabilities[pointer] < 0.01))
			std::cout << ">";
		else if (probabilities[pointer] >= 0.01) {
			if (probabilities[pointer] >= 0.03)
				std::cout << "=";
			signs = (unsigned char) ((int) (100.0 * probabilities [pointer] - 1.0)) / 2;
			for (j = 0; j < (signs - 1); j++)
				std::cout << "=";
			std::cout << ">";
		}
		std::cout << "  " << std::setprecision (2) <<  std::fixed << (double) 100.0 * probabilities [pointer] << " %" << std::endl;
	}
	std::cout << std::endl;

	std::cout << "Duration     : " << std::setprecision (2) <<  std::fixed << (long double) (finish - start) / CLOCKS_PER_SEC << " seconds." << std::endl;
	std::cout << "Seed         : " << seed << std::endl;
}
