/*
** Author: Dimitris Diochnos
** email : diochnos AT gmail DOT com
**
** File Name: mighty_classes.cpp
**
** Date Started : 20 April 2006.
** Date Finished: 28 April 2006.
**
** internals_mc (version 1.0)
** Date Started :  6 July 2007.
** Date Finished: 11 July 2007.
**
** internals_mc (version 1.1)
** Date Started : 12 July 2007.
** Date Finished: 20 July 2007.
**
**
** Tab space: 3
**
**
** The "policies" file.
**
*/


#include <iostream>
#include "sec_skills_mighty.h"
#include "mighty_classes.h"

DecisionMaker::DecisionMaker(HERO_PTR a_hero)
{
	myHero = a_hero;
}

DecisionMaker::~DecisionMaker() {}


/*
** AlwaysRight.
*/

HeroAlwaysRight::HeroAlwaysRight(HERO_PTR x):DecisionMaker(x)
{
	LoopInitializer ();
}

void HeroAlwaysRight::LoopInitializer () {}

unsigned char HeroAlwaysRight::applyPolicy (SKILL_OFFER_PTR offer)
{
	if (offer->right != NULL_SKILL)
		return 	offer->right;
	else
		return 	offer->left;
}


/*
** AlwaysLeft.
*/

HeroAlwaysLeft::HeroAlwaysLeft(HERO_PTR x):DecisionMaker(x) 
{
	LoopInitializer ();
}

void HeroAlwaysLeft::LoopInitializer () {}

unsigned char HeroAlwaysLeft::applyPolicy (SKILL_OFFER_PTR offer)
{
	if (offer->left != NULL_SKILL)
		return 	offer->left;
	else {
		std::cout << "Problem in your thoughts?" << std::endl;
		return 	NULL_SKILL;
	}
}


/*
** AlwaysLeft Then Preference.
*/

HeroAlwaysLeftThenPref::HeroAlwaysLeftThenPref(HERO_PTR x, unsigned char * selections):DecisionMaker(x)
{
	int i;

	for (i = 0; i < DIFF_SEC_SKILLS; i++)
		Preference[ selections[i] ] = i;

	LoopInitializer ();
}

void HeroAlwaysLeftThenPref::LoopInitializer () {}

unsigned char HeroAlwaysLeftThenPref::applyPolicy (SKILL_OFFER_PTR offer)
{
	if (myHero->get_skill_slot(offer->left) < SKILL_SLOTS) {
		// Skill has already been taken. Therefore we must upgrade according to policy.
		return 	offer->left;
	}
	else if (Preference[offer->left] < Preference[offer->right]) // Both skills are brand new.
		return 	offer->left;
	else // Both skills are brand new.
		return 	offer->right;
}


/*
** Seek for preferred skills otherwise upgrade.
*/

HeroSeekPrefOtherwiseUpgrade::HeroSeekPrefOtherwiseUpgrade(HERO_PTR x, unsigned char * selections):DecisionMaker(x)
{
	int i;

	for (i = 0; i < DIFF_SEC_SKILLS; i++)
		Preference[ selections[i] ] = i;

	LoopInitializer ();

	//printPreferences ();
	return;
}

void HeroSeekPrefOtherwiseUpgrade::LoopInitializer ()
{
	unwanted = 0;
}

unsigned char HeroSeekPrefOtherwiseUpgrade::applyPolicy (SKILL_OFFER_PTR offer)
{
	if ((myHero->get_skill_slot(offer->left) < SKILL_SLOTS) && (myHero->get_skill_slot(offer->right) < SKILL_SLOTS)) {
		// Both skills have already been taken. Choose the one with higher preference.
		if (Preference[offer->left] < Preference[offer->right])
			return 	offer->left;
		else
			return 	offer->right;
	}
	else if (myHero->get_skill_slot(offer->left) < SKILL_SLOTS) {
		// Left Skill ONLY has already been taken.
		//std::cout << (int) Preference[offer->right] << " (" << (int) (SKILL_SLOTS - unwanted) << ")." << std::endl;
		if ( qualifies_on_top (offer->right) )
			return 	offer->right; //Get new skill. We are sure that it's on TOP of our list AT THE MOMENT of selection. ;)
		else // offer->right is not so high on our preference list. Hence we upgrade the left one.
			return 	offer->left; // Upgrade existing skill.
	}
	else { 
		// Both skills are brand new. Pick the one you prefer more.
		if (Preference[offer->left] < Preference[offer->right]) {
			if ( !qualifies_on_top(offer->left) )
				unwanted++;
			return 	offer->left;
		}
		else {
			if ( !qualifies_on_top(offer->right) )
				unwanted++;
			return 	offer->right;
		}
	}
}

unsigned char HeroSeekPrefOtherwiseUpgrade::get_number_of_unwanted_skills () const
{
	return unwanted;
}

void 	HeroSeekPrefOtherwiseUpgrade::printPreferences ()
{
	std::cout << "\n\nPreferences [SPOU]" << std::endl;
	for (int i = 0; i < DIFF_SEC_SKILLS; i++) {
		std::cout << get_skill_name_12 ( i ) << ": " << (int) Preference[i] << std::endl;
	}
}

bool	HeroSeekPrefOtherwiseUpgrade::qualifies_on_top (const unsigned char skill) const
{
	if (Preference [skill] >= SKILL_SLOTS)
		return false;

	int i, counter;
	unsigned char sorted_pref [DIFF_SEC_SKILLS];

	for (i = 0; i < DIFF_SEC_SKILLS; i++)
		sorted_pref [ Preference [i] ] = i;

	counter = get_number_of_unwanted_skills ();
	for (i = SKILL_SLOTS - 1; i >= 0; i--) {
		if ( counter == 0 )
			break;
		if ( sorted_pref [i] == skill )
			break;
		if ( myHero->get_skill_slot ( sorted_pref [i] ) >= SKILL_SLOTS ) {
			// Out of range ==> Hero does not have this skill.
			// Moreover, this skill has lower 'priority' than given skill.
			counter--;
		}
	}

	return	(counter == 0) ? true : false;
}
